/*-
 * Copyright (c) 2011, A P Francisco
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include <assert.h>

#define LEFT(x)		((((x) + 1) << 1) - 1)
#define RIGHT(x)	 (((x) + 1) << 1)
#define PARENT(x)	((((x) + 1) >> 1) - 1)

void
binheap_exchange(size_t *heap, size_t *map, size_t i, size_t j)
{
	size_t v;

	assert(map[heap[i]] == i && map[heap[j]] == j);

	map[heap[i]] = j;
	map[heap[j]] = i;
	v = heap[i];
	heap[i] = heap[j];
	heap[j] = v;

	assert(map[heap[i]] == i && map[heap[j]] == j);
}

void
binheap_siftdown(void *key, size_t *heap, size_t *map, size_t n, size_t size,
		size_t i, int(*cmp)(const void *, const void *))
{
	size_t l = LEFT(i), r = RIGHT(i), m = i;

	if (l < n && cmp(((char *) key) + heap[l]*size,
	    ((char *) key) + heap[m]*size) < 0)
		m = l;
	if (r < n && cmp(((char *) key) + heap[r]*size,
	    ((char *) key) + heap[m]*size) < 0)
		m = r;
	if (m == i) return;
	binheap_exchange(heap, map, i, m);
	binheap_siftdown(key, heap, map, n, size, m, cmp);
}

void
binheap_siftup(void *key, size_t *heap, size_t *map, size_t n, size_t size,
		size_t i, int(*cmp)(const void *, const void *))
{
	size_t p = PARENT(i);

	if (p >= ULONG_MAX || cmp(((char *) key) + heap[p]*size,
	    ((char *) key) + heap[i]*size) <= 0) {
		return;
	}
	binheap_exchange(heap, map, i, p);
	binheap_siftup(key, heap, map, n, size, p, cmp);
}

void
binheap_make(void *key, size_t *heap, size_t *map, size_t n,
	    size_t size, int(*cmp)(const void *, const void *))
{
	int i;

	for (i = 0; i < n; i++)
		heap[i] = map[i] = i;

	for (i = PARENT(n - 1); i >= 0; i--)
		binheap_siftdown(key, heap, map, n, size, i, cmp);
}

