#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <search.h>

#include <assert.h>

#include "binheap.h"

#define MAXNAME 64
#define MAXNUM 1000000000
#define MAXPKG 5

struct call {
  unsigned long nr;
  size_t time;
  char *name, pkg;
};

static int alpha[] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '_', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'};

static struct call ** store;
static size_t storei, time;
static struct binheap * q;
static void * nameroot;
static void * nrroot;

static void do_nothing(void * n) { }

void
*genname()
{
  size_t len = rand()%MAXNAME + 1, i;
  char * name = malloc(sizeof(char)*(len+1));

  name[0] = alpha[rand()%52];
  for (i = 1; i < len; i++)
    name[i] = alpha[rand()%63];
  name[i] = '\0';

  return name;
}

unsigned long
gennum()
{
  return rand()%MAXNUM;
}

char
genpkg()
{
  return 'A' + rand()%MAXPKG;
}

int
callcmp(const void *ap, const void *bp)
{
	struct call * a = * (struct call **) ap;
	struct call * b = * (struct call **) bp;

	if (a->pkg < b->pkg)
		return -1;
	if (a->pkg > b->pkg)
		return 1;

  return a->time - b->time;
}

static int
nrcmp (const void *pa, const void *pb)
{
  unsigned long la = (unsigned long) pa;
  unsigned long lb = (unsigned long) pb;

  if (la < lb)
    return -1;
  if (la > lb)
    return 1;
  return 0;
}

static int
namecmp (const void *pa, const void *pb)
{
  return strcmp((char *) pa, (char *) pb);
}

int
main(int argc, char * argv[])
{
  struct call * c;
  int i, k, n, op, maxop;
  char pkg;

  if (argc != 4) {
    fprintf(stderr, "Usage: %s <seed> <size> <maxop>\nseed\tsemente para o srand\nsize\ttamanho do teste\nmaxop\toperacoes a considerar, valores possiveis: 1, 2, 3, 4, 5\n", argv[0]);
    return EXIT_FAILURE;
  }
  srand(atoi(argv[1]));
  k = atoi(argv[2]);
  maxop = atoi(argv[3]);

  store = malloc(sizeof(struct call *)*2*k);
  memset(store, 0x0, sizeof(struct call *)*2*k);
  storei = time = 0;
  nameroot = nrroot = NULL;

  for (i = 0; i < k; i++) {
    c = malloc(sizeof(struct call));
    c->name = genname();
    while (tfind(c->name, &nameroot, namecmp)) {
      free(c->name);
      c->name = genname();
    }
    tsearch(c->name, &nameroot, namecmp);

    c->nr = gennum();
    while (tfind((void *) c->nr, &nrroot, nrcmp))
      c->nr = gennum();
    tsearch((void *) c->nr, &nrroot, nrcmp);

    c->pkg = genpkg();
    c->time = time ++;
    printf("a %s %lu %c\n", c->name, c->nr, c->pkg);
    store[storei++] = c;
  }

  assert(k == storei);

  BINHEAP_INIT(q, struct call *, store, 2*k, callcmp);
  BINHEAP_MAKE(q, k);
  assert(q->size == storei);

  for (i = 0; i < k; i++) {
    op = rand()%maxop;
    
    switch (op) {
      case 0:
        c = malloc(sizeof(struct call));
        c->name = genname();
        while (tfind(c->name, &nameroot, namecmp)) {
          free(c->name);
          c->name = genname();
        }
        tsearch(c->name, &nameroot, namecmp);

        c->nr = gennum();
        while (tfind((void *) c->nr, &nrroot, nrcmp))
          c->nr = gennum();
        tsearch((void *) c->nr, &nrroot, nrcmp);

        c->pkg = genpkg();
        c->time = time ++;
        printf("a %s %lu %c\n", c->name, c->nr, c->pkg);
        store[storei++] = c;

        BINHEAP_UPDATE(q, storei-1);
        assert(q->size == storei);
        break;
      case 1:
        printf("l\n");
        break;
      case 2:
        printf("p\n");
        n = BINHEAP_TOP(q);
        BINHEAP_POP(q);
        /* printf("\t%s %lu %c (%lu)\n", store[n]->name, store[n]->nr, store[n]->pkg, store[n]->time); */
        tdelete(store[n]->name, &nameroot, namecmp);
        tdelete((void *) store[n]->nr, &nrroot, nrcmp);
        free(store[n]->name);
        free(store[n]);
        store[n] = NULL;
        storei--;
        if (storei != n) {
          BINHEAP_DEL(q, storei);
          store[n] = store[storei];
          store[storei] = NULL;
          BINHEAP_UPDATE(q, n);
        }
        assert(q->size == storei);
        break;
      case 3:
        n = rand()%storei;
        pkg = genpkg();
        if (pkg != store[n]->pkg) {
          store[n]->pkg = pkg;
          store[n]->time = time ++;
          printf("u %lu %c\n", store[n]->nr, store[n]->pkg);
          BINHEAP_UPDATE(q, n);
          assert(q->size == storei);
        }
        break;
      case 4:
        n = rand()%storei;
        printf("r %s\n", store[n]->name);
        BINHEAP_DEL(q, n);
        tdelete(store[n]->name, &nameroot, namecmp);
        tdelete((void *) store[n]->nr, &nrroot, nrcmp);
        free(store[n]->name);
        free(store[n]);
        storei--;
        if (storei != n) {
          BINHEAP_DEL(q, storei);
          store[n] = store[storei];
          store[storei] = NULL;
          BINHEAP_UPDATE(q, n);
        }
        assert(q->size == storei);
        break;
    }
  }

  printf("x\n");

  BINHEAP_FREE(q);
  for (i = 0; i < storei; i++) {
    free(store[i]->name);
    free(store[i]);
  }
  free(store);
  tdestroy(nameroot, do_nothing);
  tdestroy(nrroot, do_nothing);

  return EXIT_SUCCESS;
}

